@@ -50,6 +50,8 @@ static int whichreg(long regflags, int regval) {
 	0, 0, 0, R_TR3, R_TR4, R_TR5, R_TR6, R_TR7 };
     static int fpureg[] = {
 	R_ST0, R_ST1, R_ST2, R_ST3, R_ST4, R_ST5, R_ST6, R_ST7 };
+    static int mmxreg[] = {
+	R_MM0, R_MM1, R_MM2, R_MM3, R_MM4, R_MM5, R_MM6, R_MM7 };
 
     if (!(REG_AL & ~regflags))
 	return R_AL;
@@ -85,6 +87,8 @@ static int whichreg(long regflags, int regval) {
 	return treg[regval];
     if (!(FPUREG & ~regflags))
 	return fpureg[regval];
+    if (!(MMXREG & ~regflags))
+	return mmxreg[regval];
     return 0;
 }
 
@@ -226,8 +230,8 @@ static unsigned char *do_ea (unsigned char *data, int modrm, int asize,
 	    op->segment |= SEG_DISP32;
 	    op->offset = *data++;
 	    op->offset |= (*data++) << 8;
-	    op->offset |= (*data++) << 16;
-	    op->offset |= (*data++) << 24;
+	    op->offset |= ((long) *data++) << 16;
+	    op->offset |= ((long) *data++) << 24;
 	    break;
 	}
 	return data;
@@ -241,6 +245,7 @@ static unsigned char *do_ea (unsigned char *data, int modrm, int asize,
 static int matches (unsigned char *r, unsigned char *data, int asize,
 		    int osize, int segsize, insn *ins) {
     unsigned char *origdata = data;
+    int a_used = FALSE, o_used = FALSE;
 
     while (*r) {
 	int c = *r++;
@@ -308,8 +313,8 @@ static int matches (unsigned char *r, unsigned char *data, int asize,
 	    ins->oprs[c-034].offset = *data++;
 	    ins->oprs[c-034].offset |= (*data++ << 8);
 	    if (asize == 32) {
-		ins->oprs[c-034].offset |= (*data++ << 16);
-		ins->oprs[c-034].offset |= (*data++ << 24);
+		ins->oprs[c-034].offset |= (((long) *data++) << 16);
+		ins->oprs[c-034].offset |= (((long) *data++) << 24);
 	    }
 	    if (segsize != asize)
 		ins->oprs[c-034].addr_size = asize;
@@ -317,8 +322,8 @@ static int matches (unsigned char *r, unsigned char *data, int asize,
 	if (c >= 040 && c <= 042) {
 	    ins->oprs[c-040].offset = *data++;
 	    ins->oprs[c-040].offset |= (*data++ << 8);
-	    ins->oprs[c-040].offset |= (*data++ << 16);
-	    ins->oprs[c-040].offset |= (*data++ << 24);
+	    ins->oprs[c-040].offset |= (((long) *data++) << 16);
+	    ins->oprs[c-040].offset |= (((long) *data++) << 24);
 	}
 	if (c >= 050 && c <= 052) {
 	    ins->oprs[c-050].offset = (signed char) *data++;
@@ -334,8 +339,8 @@ static int matches (unsigned char *r, unsigned char *data, int asize,
 	    ins->oprs[c-064].offset = *data++;
 	    ins->oprs[c-064].offset |= (*data++ << 8);
 	    if (asize == 32) {
-		ins->oprs[c-064].offset |= (*data++ << 16);
-		ins->oprs[c-064].offset |= (*data++ << 24);
+		ins->oprs[c-064].offset |= (((long) *data++) << 16);
+		ins->oprs[c-064].offset |= (((long) *data++) << 24);
 		ins->oprs[c-064].segment |= SEG_32BIT;
 	    } else
 		ins->oprs[c-064].segment &= ~SEG_32BIT;
@@ -346,8 +351,8 @@ static int matches (unsigned char *r, unsigned char *data, int asize,
 	if (c >= 070 && c <= 072) {
 	    ins->oprs[c-070].offset = *data++;
 	    ins->oprs[c-070].offset |= (*data++ << 8);
-	    ins->oprs[c-070].offset |= (*data++ << 16);
-	    ins->oprs[c-070].offset |= (*data++ << 24);
+	    ins->oprs[c-070].offset |= (((long) *data++) << 16);
+	    ins->oprs[c-070].offset |= (((long) *data++) << 24);
 	    ins->oprs[c-070].segment |= SEG_32BIT | SEG_RELATIVE;
 	}
 	if (c >= 0100 && c <= 0177) {
@@ -369,13 +374,44 @@ static int matches (unsigned char *r, unsigned char *data, int asize,
 		ins->oprs[c-0300].segment |= SEG_32BIT;
 	    else
 		ins->oprs[c-0300].segment &= ~SEG_32BIT;
+	    a_used = TRUE;
+	}
+	if (c == 0310) {
+	    if (asize == 32)
+		return FALSE;
+	    else
+		a_used = TRUE;
+	}
+	if (c == 0311) {
+	    if (asize == 16)
+		return FALSE;
+	    else
+		a_used = TRUE;
+	}
+	if (c == 0312) {
+	    if (asize != segsize)
+		return FALSE;
+	    else
+		a_used = TRUE;
+	}
+	if (c == 0320) {
+	    if (osize == 32)
+		return FALSE;
+	    else
+		o_used = TRUE;
+	}
+	if (c == 0321) {
+	    if (osize == 16)
+		return FALSE;
+	    else
+		o_used = TRUE;
+	}
+	if (c == 0322) {
+	    if (osize != segsize)
+		return FALSE;
+	    else
+		o_used = TRUE;
 	}
-	if (c == 0310 && asize == 32) return FALSE;
-	if (c == 0311 && asize == 16) return FALSE;
-	if (c == 0312 && asize != segsize) return FALSE;
-	if (c == 0320 && osize == 32) return FALSE;
-	if (c == 0321 && osize == 16) return FALSE;
-	if (c == 0322 && osize != segsize) return FALSE;
 	if (c == 0330) {
 	    int t = *r++, d = *data++;
 	    if (d < t || d > t+15)
@@ -385,6 +421,15 @@ static int matches (unsigned char *r, unsigned char *data, int asize,
 	}
     }
 
+    /*
+     * Check for unused a/o prefixes.
+     */
+    ins->nprefix = 0;
+    if (!a_used && asize != segsize)
+	ins->prefixes[ins->nprefix++] = (asize == 16 ? P_A16 : P_A32);
+    if (!o_used && osize != segsize)
+	ins->prefixes[ins->nprefix++] = (osize == 16 ? P_O16 : P_O32);
+
     return data - origdata;
 }
 
@@ -395,6 +440,7 @@ long disasm (unsigned char *data, char *output, int segsize, long offset,
     char *segover;
     int rep, lock, asize, osize, i, slen, colon;
     unsigned char *origdata;
+    int works;
     insn ins;
 
     /*
@@ -431,24 +477,25 @@ long disasm (unsigned char *data, char *output, int segsize, long offset,
     ins.oprs[0].addr_size = ins.oprs[1].addr_size = ins.oprs[2].addr_size =
 	(segsize == 16 ? 0 : SEG_32BIT);
     ins.condition = -1;
+    works = TRUE;
     for (p = itable[*data]; *p; p++)
 	if ( (length = matches((unsigned char *)((*p)->code), data,
 			       asize, osize, segsize, &ins)) ) {
-	    int works = TRUE;
+	    works = TRUE;
 	    /*
 	     * Final check to make sure the types of r/m match up.
 	     */
 	    for (i = 0; i < (*p)->operands; i++)
 		if (((ins.oprs[i].segment & SEG_RMREG) &&
-		     !(MEMORY & ~(*p)->operand[i])) ||
+		     !(MEMORY & ~(*p)->opd[i])) ||
 		    (!(ins.oprs[i].segment & SEG_RMREG) &&
-		     !(REGNORM & ~(*p)->operand[i]) &&
-		     !((*p)->operand[i] & REG_SMASK)))
+		     !(REGNORM & ~(*p)->opd[i]) &&
+		     !((*p)->opd[i] & REG_SMASK)))
 		    works = FALSE;
 	    if (works)
 		break;
 	}
-    if (!length)
+    if (!length || !works)
 	return 0;		       /* no instruction was matched */
 
     slen = 0;
@@ -465,6 +512,13 @@ long disasm (unsigned char *data, char *output, int segsize, long offset,
     }
     if (lock)
 	slen += sprintf(output+slen, "lock ");
+    for (i = 0; i < ins.nprefix; i++)
+	switch (ins.prefixes[i]) {
+	  case P_A16: slen += sprintf(output+slen, "a16 "); break;
+	  case P_A32: slen += sprintf(output+slen, "a32 "); break;
+	  case P_O16: slen += sprintf(output+slen, "o16 "); break;
+	  case P_O32: slen += sprintf(output+slen, "o32 "); break;
+	}
 
     for (i = 0; i < elements(ico); i++)
 	if ((*p)->opcode == ico[i]) {
@@ -490,24 +544,24 @@ long disasm (unsigned char *data, char *output, int segsize, long offset,
 	     * add sync marker, if autosync is on
 	     */
 	    if (autosync)
-		add_sync (ins.oprs[i].offset);
+		add_sync (ins.oprs[i].offset, 0L);
 	}
 
-	if ((*p)->operand[i] & COLON)
+	if ((*p)->opd[i] & COLON)
 	    colon = TRUE;
 	else
 	    colon = FALSE;
 
-	if (((*p)->operand[i] & (REGISTER | FPUREG)) ||
+	if (((*p)->opd[i] & (REGISTER | FPUREG)) ||
 	    (ins.oprs[i].segment & SEG_RMREG)) {
-	    ins.oprs[i].basereg = whichreg ((*p)->operand[i],
+	    ins.oprs[i].basereg = whichreg ((*p)->opd[i],
 					    ins.oprs[i].basereg);
 	    slen += sprintf(output+slen, "%s",
 			    reg_names[ins.oprs[i].basereg]);
-	} else if (!(UNITY & ~(*p)->operand[i])) {
+	} else if (!(UNITY & ~(*p)->opd[i])) {
 	    output[slen++] = '1';
-	} else if ( (*p)->operand[i] & IMMEDIATE ) {
-	    if ( (*p)->operand[i] & BITS8 ) {
+	} else if ( (*p)->opd[i] & IMMEDIATE ) {
+	    if ( (*p)->opd[i] & BITS8 ) {
 		slen += sprintf(output+slen, "byte ");
 		if (ins.oprs[i].segment & SEG_SIGNED) {
 		    if (ins.oprs[i].offset < 0) {
@@ -516,17 +570,17 @@ long disasm (unsigned char *data, char *output, int segsize, long offset,
 		    } else
 			output[slen++] = '+';
 		}
-	    } else if ( (*p)->operand[i] & BITS16 ) {
+	    } else if ( (*p)->opd[i] & BITS16 ) {
 		slen += sprintf(output+slen, "word ");
-	    } else if ( (*p)->operand[i] & BITS32 ) {
+	    } else if ( (*p)->opd[i] & BITS32 ) {
 		slen += sprintf(output+slen, "dword ");
-	    } else if ( (*p)->operand[i] & NEAR ) {
+	    } else if ( (*p)->opd[i] & NEAR ) {
 		slen += sprintf(output+slen, "near ");
-	    } else if ( (*p)->operand[i] & SHORT ) {
+	    } else if ( (*p)->opd[i] & SHORT ) {
 		slen += sprintf(output+slen, "short ");
 	    }
 	    slen += sprintf(output+slen, "0x%lx", ins.oprs[i].offset);
-	} else if ( !(MEM_OFFS & ~(*p)->operand[i]) ) {
+	} else if ( !(MEM_OFFS & ~(*p)->opd[i]) ) {
 	    slen += sprintf(output+slen, "[%s%s%s0x%lx]",
 			    (segover ? segover : ""),
 			    (segover ? ":" : ""),
@@ -534,21 +588,21 @@ long disasm (unsigned char *data, char *output, int segsize, long offset,
 			     ins.oprs[i].addr_size == 16 ? "word " : ""),
 			    ins.oprs[i].offset);
 	    segover = NULL;
-	} else if ( !(REGMEM & ~(*p)->operand[i]) ) {
+	} else if ( !(REGMEM & ~(*p)->opd[i]) ) {
 	    int started = FALSE;
-	    if ( (*p)->operand[i] & BITS8 )
+	    if ( (*p)->opd[i] & BITS8 )
 		slen += sprintf(output+slen, "byte ");
-	    if ( (*p)->operand[i] & BITS16 )
+	    if ( (*p)->opd[i] & BITS16 )
 		slen += sprintf(output+slen, "word ");
-	    if ( (*p)->operand[i] & BITS32 )
+	    if ( (*p)->opd[i] & BITS32 )
 		slen += sprintf(output+slen, "dword ");
-	    if ( (*p)->operand[i] & BITS64 )
+	    if ( (*p)->opd[i] & BITS64 )
 		slen += sprintf(output+slen, "qword ");
-	    if ( (*p)->operand[i] & BITS80 )
+	    if ( (*p)->opd[i] & BITS80 )
 		slen += sprintf(output+slen, "tword ");
-	    if ( (*p)->operand[i] & FAR )
+	    if ( (*p)->opd[i] & FAR )
 		slen += sprintf(output+slen, "far ");
-	    if ( (*p)->operand[i] & NEAR )
+	    if ( (*p)->opd[i] & NEAR )
 		slen += sprintf(output+slen, "near ");
 	    output[slen++] = '[';
 	    if (ins.oprs[i].addr_size)
@@ -597,7 +651,10 @@ long disasm (unsigned char *data, char *output, int segsize, long offset,
     }
     output[slen] = '\0';
     if (segover) {		       /* unused segment override */
-	memmove (output+3, output, slen+1);
+	char *p = output;
+	int count = slen+1;
+	while (count--)
+	    p[count+3] = p[count];
 	strncpy (output, segover, 2);
 	output[2] = ' ';
     }
