@@ -11,7 +11,9 @@
 #ifndef NASM_H
 #define NASM_H
 
-#define NASM_VER "0.90"
+#define NASM_MAJOR_VER 0
+#define NASM_MINOR_VER 93
+#define NASM_VER "0.93"
 
 #ifndef NULL
 #define NULL 0
@@ -48,7 +50,7 @@ struct ofmt;
  */
 typedef void (*efunc) (int severity, char *fmt, ...);
 
-/* 
+/*
  * These are the error severity codes which get passed as the first
  * argument to an efunc.
  */
@@ -123,6 +125,7 @@ typedef void (*ldfunc) (char *label, long segment, long offset,
 #define REG32     0x00201004L
 #define FPUREG    0x01000000L	       /* floating point stack registers */
 #define FPU0      0x01000800L	       /* FPU stack register zero */
+#define MMXREG    0x00001008L	       /* MMX registers */
 
 /* special register operands: these may be treated differently */
 #define REG_SMASK 0x00070000L	       /* a mask for the following */
@@ -160,7 +163,8 @@ enum {				       /* register names */
     R_AH = 1, R_AL, R_AX, R_BH, R_BL, R_BP, R_BX, R_CH, R_CL, R_CR0,
     R_CR2, R_CR3, R_CR4, R_CS, R_CX, R_DH, R_DI, R_DL, R_DR0, R_DR1,
     R_DR2, R_DR3, R_DR6, R_DR7, R_DS, R_DX, R_EAX, R_EBP, R_EBX,
-    R_ECX, R_EDI, R_EDX, R_ES, R_ESI, R_ESP, R_FS, R_GS, R_SI, R_SP,
+    R_ECX, R_EDI, R_EDX, R_ES, R_ESI, R_ESP, R_FS, R_GS, R_MM0,
+    R_MM1, R_MM2, R_MM3, R_MM4, R_MM5, R_MM6, R_MM7, R_SI, R_SP,
     R_SS, R_ST0, R_ST1, R_ST2, R_ST3, R_ST4, R_ST5, R_ST6, R_ST7,
     R_TR3, R_TR4, R_TR5, R_TR6, R_TR7, REG_ENUM_LIMIT
 };
@@ -171,10 +175,10 @@ enum {				       /* instruction names */
     I_CALL, I_CBW, I_CDQ, I_CLC, I_CLD, I_CLI, I_CLTS, I_CMC, I_CMP,
     I_CMPSB, I_CMPSD, I_CMPSW, I_CMPXCHG, I_CMPXCHG8B, I_CPUID,
     I_CWD, I_CWDE, I_DAA, I_DAS, I_DB, I_DD, I_DEC, I_DIV, I_DQ,
-    I_DT, I_DW, I_ENTER, I_EQU, I_F2XM1, I_FABS, I_FADD, I_FADDP,
-    I_FBLD, I_FBSTP, I_FCHS, I_FCLEX, I_FCMOVB, I_FCMOVBE, I_FCMOVE,
-    I_FCMOVNB, I_FCMOVNBE, I_FCMOVNE, I_FCMOVNU, I_FCMOVU, I_FCOM,
-    I_FCOMI, I_FCOMIP, I_FCOMP, I_FCOMPP, I_FCOS, I_FDECSTP,
+    I_DT, I_DW, I_EMMS, I_ENTER, I_EQU, I_F2XM1, I_FABS, I_FADD,
+    I_FADDP, I_FBLD, I_FBSTP, I_FCHS, I_FCLEX, I_FCMOVB, I_FCMOVBE,
+    I_FCMOVE, I_FCMOVNB, I_FCMOVNBE, I_FCMOVNE, I_FCMOVNU, I_FCMOVU,
+    I_FCOM, I_FCOMI, I_FCOMIP, I_FCOMP, I_FCOMPP, I_FCOS, I_FDECSTP,
     I_FDISI, I_FDIV, I_FDIVP, I_FDIVR, I_FDIVRP, I_FENI, I_FFREE,
     I_FIADD, I_FICOM, I_FICOMP, I_FIDIV, I_FIDIVR, I_FILD, I_FIMUL,
     I_FINCSTP, I_FINIT, I_FIST, I_FISTP, I_FISUB, I_FISUBR, I_FLD,
@@ -190,18 +194,26 @@ enum {				       /* instruction names */
     I_IRETD, I_IRETW, I_JCXZ, I_JECXZ, I_JMP, I_LAHF, I_LAR, I_LDS,
     I_LEA, I_LEAVE, I_LES, I_LFS, I_LGDT, I_LGS, I_LIDT, I_LLDT,
     I_LMSW, I_LOADALL, I_LODSB, I_LODSD, I_LODSW, I_LOOP, I_LOOPE,
-    I_LOOPNE, I_LOOPNZ, I_LOOPZ, I_LSL, I_LSS, I_LTR, I_MOV,
-    I_MOVSB, I_MOVSD, I_MOVSW, I_MOVSX, I_MOVZX, I_MUL, I_NEG,
-    I_NOP, I_NOT, I_OR, I_OUT, I_OUTSB, I_OUTSD, I_OUTSW, I_POP,
-    I_POPA, I_POPAD, I_POPAW, I_POPF, I_POPFD, I_POPFW, I_PUSH,
-    I_PUSHA, I_PUSHAD, I_PUSHAW, I_PUSHF, I_PUSHFD, I_PUSHFW, I_RCL,
-    I_RCR, I_RDMSR, I_RDPMC, I_RDTSC, I_RESB, I_RESD, I_RESQ,
-    I_REST, I_RESW, I_RET, I_RETF, I_RETN, I_ROL, I_ROR, I_RSM,
-    I_SAHF, I_SAL, I_SALC, I_SAR, I_SBB, I_SCASB, I_SCASD, I_SCASW,
-    I_SGDT, I_SHL, I_SHLD, I_SHR, I_SHRD, I_SIDT, I_SLDT, I_SMSW,
-    I_STC, I_STD, I_STI, I_STOSB, I_STOSD, I_STOSW, I_STR, I_SUB,
-    I_TEST, I_UMOV, I_VERR, I_VERW, I_WAIT, I_WBINVD, I_WRMSR,
-    I_XADD, I_XCHG, I_XLATB, I_XOR, I_CMOVcc, I_Jcc, I_SETcc
+    I_LOOPNE, I_LOOPNZ, I_LOOPZ, I_LSL, I_LSS, I_LTR, I_MOV, I_MOVD,
+    I_MOVQ, I_MOVSB, I_MOVSD, I_MOVSW, I_MOVSX, I_MOVZX, I_MUL,
+    I_NEG, I_NOP, I_NOT, I_OR, I_OUT, I_OUTSB, I_OUTSD, I_OUTSW,
+    I_PACKSSDW, I_PACKSSWB, I_PACKUSWB, I_PADDB, I_PADDD, I_PADDSB,
+    I_PADDSW, I_PADDUSB, I_PADDUSW, I_PADDW, I_PAND, I_PANDN,
+    I_PCMPEQB, I_PCMPEQD, I_PCMPEQW, I_PCMPGTB, I_PCMPGTD,
+    I_PCMPGTW, I_PMADDWD, I_PMULHW, I_PMULLW, I_POP, I_POPA,
+    I_POPAD, I_POPAW, I_POPF, I_POPFD, I_POPFW, I_POR, I_PSLLD,
+    I_PSLLQ, I_PSLLW, I_PSRAD, I_PSRAW, I_PSRLD, I_PSRLQ, I_PSRLW,
+    I_PSUBB, I_PSUBD, I_PSUBSB, I_PSUBSW, I_PSUBUSB, I_PSUBUSW,
+    I_PSUBW, I_PUNPCKHBW, I_PUNPCKHDQ, I_PUNPCKHWD, I_PUNPCKLBW,
+    I_PUNPCKLDQ, I_PUNPCKLWD, I_PUSH, I_PUSHA, I_PUSHAD, I_PUSHAW,
+    I_PUSHF, I_PUSHFD, I_PUSHFW, I_PXOR, I_RCL, I_RCR, I_RDMSR,
+    I_RDPMC, I_RDTSC, I_RESB, I_RESD, I_RESQ, I_REST, I_RESW, I_RET,
+    I_RETF, I_RETN, I_ROL, I_ROR, I_RSM, I_SAHF, I_SAL, I_SALC,
+    I_SAR, I_SBB, I_SCASB, I_SCASD, I_SCASW, I_SGDT, I_SHL, I_SHLD,
+    I_SHR, I_SHRD, I_SIDT, I_SLDT, I_SMSW, I_STC, I_STD, I_STI,
+    I_STOSB, I_STOSD, I_STOSW, I_STR, I_SUB, I_TEST, I_UMOV, I_VERR,
+    I_VERW, I_WAIT, I_WBINVD, I_WRMSR, I_XADD, I_XCHG, I_XLATB,
+    I_XOR, I_CMOVcc, I_Jcc, I_SETcc
 };
 
 enum {				       /* condition code names */
@@ -217,8 +229,8 @@ enum {				       /* condition code names */
  */
 enum {				       /* instruction prefixes */
     PREFIX_ENUM_START = REG_ENUM_LIMIT,
-    P_LOCK = PREFIX_ENUM_START, P_REP, P_REPE, P_REPNE, P_REPNZ, P_REPZ,
-    P_TIMES
+    P_A16 = PREFIX_ENUM_START, P_A32, P_LOCK, P_O16, P_O32, P_REP, P_REPE,
+    P_REPNE, P_REPNZ, P_REPZ, P_TIMES
 };
 
 enum {				       /* extended operand types */
@@ -244,9 +256,11 @@ typedef struct extop {		       /* extended operand */
     long wrt;			       /* ... and here */
 } extop;
 
+#define MAXPREFIX 4
+
 typedef struct {		       /* an instruction itself */
     char *label;		       /* the label defined, or NULL */
-    int prefixes[2];		       /* instruction prefixes, if any */
+    int prefixes[MAXPREFIX];	       /* instruction prefixes, if any */
     int nprefix;		       /* number of entries in above */
     int opcode;			       /* the opcode - not just the string */
     int condition;		       /* the condition code, if Jcc/SETcc */
@@ -254,6 +268,7 @@ typedef struct {		       /* an instruction itself */
     operand oprs[3];	   	       /* the operands, defined as above */
     extop *eops;		       /* extended operands */
     int times;			       /* repeat count (TIMES prefix) */
+    int forw_ref;		       /* is there a forward reference? */
 } insn;
 
 /*
@@ -322,10 +337,14 @@ struct ofmt {
      *
      * It may also be called with NULL, in which case it is to
      * return the _default_ section number for starting assembly in.
-     * 
+     *
      * It is allowed to modify the string it is given a pointer to.
+     *
+     * It is also allowed to specify a default instruction size for
+     * the segment, by setting `*bits' to 16 or 32. Or, if it
+     * doesn't wish to define a default, it can leave `bits' alone.
      */
-    long (*section) (char *name, int pass);
+    long (*section) (char *name, int pass, int *bits);
 
     /*
      * This procedure is called to modify the segment base values
@@ -343,7 +362,7 @@ struct ofmt {
      * directive word in `directive' and the parameter string in
      * `value'. It is called in both assembly passes, and `pass'
      * will be either 1 or 2.
-     * 
+     *
      * This procedure should return zero if it does not _recognise_
      * the directive, so that the main program can report an error.
      * If it recognises the directive but then has its own errors,
@@ -376,7 +395,7 @@ struct ofmt {
      * the output driver to clean itself up and free its memory.
      * Typically, it will also be the point at which the object
      * file actually gets _written_.
-     * 
+     *
      * One thing the cleanup routine should always do is to close
      * the output file pointer.
      */
