@@ -15,6 +15,9 @@
 
 #include "nasm.h"
 #include "nasmlib.h"
+#include "outform.h"
+
+#if defined(OF_COFF) || defined(OF_WIN32)
 
 /*
  * Notes on COFF:
@@ -30,10 +33,10 @@
  * to add the address of the symbol and then subtract the address
  * of THE BYTE AFTER THE RELOCATED DWORD. Hence the two formats are
  * subtly incompatible.
- * 
+ *
  * (2) Win32 doesn't bother putting any flags in the header flags
  * field (at offset 0x12 into the file).
- * 
+ *
  * (3) Win32 puts some weird flags into the section header table.
  * It uses flags 0x80000000 (writable), 0x40000000 (readable) and
  * 0x20000000 (executable) in the expected combinations, which
@@ -52,7 +55,7 @@
  * seem to fit at all. In fact, Win32 even includes non-linked
  * sections such as .drectve in this calculation. Not that I can be
  * bothered with those things anyway.
- * 
+ *
  * (5) Standard COFF does something very strange to common
  * variables: the relocation point for a common variable is as far
  * _before_ the variable as its size stretches out _after_ it. So
@@ -67,6 +70,11 @@ struct Reloc {
     struct Reloc *next;
     long address;		       /* relative to _start_ of section */
     long symbol;		       /* symbol number */
+    enum {
+	SECT_SYMBOLS,
+	ABS_SYMBOL,
+	REAL_SYMBOLS
+    } symbase;			       /* relocation for symbol number :) */
     int relative;		       /* TRUE or FALSE */
 };
 
@@ -84,52 +92,36 @@ static efunc error;
 static char coff_infile[FILENAME_MAX];
 
 struct Section {
-    unsigned char *data;
-    unsigned long len, size;
+    struct SAA *data;
+    unsigned long len;
     int nrelocs;
     long index;
     struct Reloc *head, **tail;
+    unsigned long flags;	       /* section flags */
+    char name[9];
+    long pos, relpos;
 };
 
-static struct Section stext, sdata;
-static unsigned long bsslen;
-static long bssindex;
+#define TEXT_FLAGS (win32 ? 0x60500020L : 0x20L)
+#define DATA_FLAGS (win32 ? 0xC0300040L : 0x40L)
+#define BSS_FLAGS (win32 ? 0xC0300080L : 0x80L)
+#define INFO_FLAGS 0x00100A00L
 
-static struct Symbol *syms;
-static unsigned long nsyms, symsize;
+#define SECT_DELTA 32
+static struct Section **sects;
+static int nsects, sectlen;
 
-static unsigned long *bsym;
-static unsigned long bsymsize;
+static struct SAA *syms;
+static unsigned long nsyms;
 
-static char *strs;
-static unsigned long strslen, strsize;
+static long def_seg;
 
-#define SECT_DELTA 4096
-#define STRTAB_DELTA 4096
-#define SYMTAB_DELTA 256
-#define BSYM_DELTA 128
+static int initsym;
 
-/*
- * The symbol table contains a double entry for the file name, a
- * double entry for each of the three sections, and an absolute
- * symbol referencing address zero, followed by the _real_ symbols.
- * That's nine extra symbols.
- */
-#define SYM_INITIAL 9
+static struct RAA *bsym, *symval;
 
-/*
- * Symbol table indices we can relocate relative to.
- */
-#define SYM_ABS_SEG 8
-#define SYM_TEXT_SEG 2
-#define SYM_DATA_SEG 4
-#define SYM_BSS_SEG 6
-
-/*
- * The section header table ends at this offset: 0x14 for the
- * header, plus 0x28 for each of three sections.
- */
-#define COFF_HDRS_END 0x8c
+static struct SAA *strs;
+static unsigned long strslen;
 
 static void coff_gen_init(FILE *, efunc);
 static void coff_sect_write (struct Section *, unsigned char *,
@@ -154,112 +146,188 @@ static void coff_std_init(FILE *fp, efunc errfunc, ldfunc ldef) {
 static void coff_gen_init(FILE *fp, efunc errfunc) {
     coffp = fp;
     error = errfunc;
-    stext.data = NULL; stext.head = NULL; stext.tail = &stext.head;
-    sdata.data = NULL; sdata.head = NULL; sdata.tail = &sdata.head;
-    stext.len = stext.size = sdata.len = sdata.size = bsslen = 0;
-    stext.nrelocs = sdata.nrelocs = 0;
-    stext.index = seg_alloc(); 
-    sdata.index = seg_alloc();
-    bssindex = seg_alloc();
-    syms = NULL;
-    nsyms = symsize = 0;
-    bsym = NULL;
-    bsymsize = 0;
-    strs = NULL;
-    strslen = strsize = 0;    
+    sects = NULL;
+    nsects = sectlen = 0;
+    syms = saa_init((long)sizeof(struct Symbol));
+    nsyms = 0;
+    bsym = raa_init();
+    symval = raa_init();
+    strs = saa_init(1L);
+    strslen = 0;
+    def_seg = seg_alloc();
 }
 
 static void coff_cleanup(void) {
     struct Reloc *r;
+    int i;
 
     coff_write();
     fclose (coffp);
-    nasm_free (stext.data);
-    while (stext.head) {
-	r = stext.head;
-	stext.head = stext.head->next;
-	nasm_free (r);
+    for (i=0; i<nsects; i++) {
+	if (sects[i]->data)
+	    saa_free (sects[i]->data);
+	while (sects[i]->head) {
+	    r = sects[i]->head;
+	    sects[i]->head = sects[i]->head->next;
+	    nasm_free (r);
+	}
     }
-    nasm_free (sdata.data);
-    while (sdata.head) {
-	r = sdata.head;
-	sdata.head = sdata.head->next;
-	nasm_free (r);
-    }
-    nasm_free (syms); 
-    nasm_free (bsym);
-    nasm_free (strs);
+    nasm_free (sects);
+    saa_free (syms);
+    raa_free (bsym);
+    raa_free (symval);
+    saa_free (strs);
 }
 
-static long coff_section_names (char *name, int pass) {
-    if (!name)
-	return stext.index;
+static int coff_make_section (char *name, unsigned long flags) {
+    struct Section *s;
 
-    if (!strcmp(name, ".text"))
-	return stext.index;
-    else if (!strcmp(name, ".data"))
-	return sdata.index;
-    else if (!strcmp(name, ".bss"))
-	return bssindex;
+    s = nasm_malloc (sizeof(*s));
+
+    if (flags != BSS_FLAGS)
+	s->data = saa_init (1L);
     else
-	return NO_SEG;
+	s->data = NULL;
+    s->head = NULL;
+    s->tail = &s->head;
+    s->len = 0;
+    s->nrelocs = 0;
+    if (!strcmp(name, ".text"))
+	s->index = def_seg;
+    else
+	s->index = seg_alloc();
+    strncpy (s->name, name, 8);
+    s->name[8] = '\0';
+    s->flags = flags;
+
+    if (nsects >= sectlen)
+	sects = nasm_realloc (sects, (sectlen += SECT_DELTA)*sizeof(*sects));
+    sects[nsects++] = s;
+
+    return nsects-1;
+}
+
+static long coff_section_names (char *name, int pass, int *bits) {
+    char *p;
+    unsigned long flags;
+    int i;
+
+    /*
+     * Default is 32 bits.
+     */
+    if (!name)
+	*bits = 32;
+
+    if (!name)
+	return def_seg;
+
+    p = name;
+    while (*p && !isspace(*p)) p++;
+    if (*p) *p++ = '\0';
+    if (strlen(p) > 8) {
+	error (ERR_WARNING, "COFF section names limited to 8 characters:"
+	       " truncating");
+	p[8] = '\0';
+    }
+    flags = 0;
+
+    while (*p && isspace(*p)) p++;
+    while (*p) {
+	char *q = p;
+	while (*p && !isspace(*p)) p++;
+	if (*p) *p++ = '\0';
+	while (*p && isspace(*p)) p++;
+	
+	if (!nasm_stricmp(q, "code") || !nasm_stricmp(q, "text")) {
+	    flags = TEXT_FLAGS;
+	} else if (!nasm_stricmp(q, "data")) {
+	    flags = DATA_FLAGS;
+	} else if (!nasm_stricmp(q, "bss")) {
+	    flags = BSS_FLAGS;
+	} else if (!nasm_stricmp(q, "info")) {
+	    if (win32)
+		flags = INFO_FLAGS;
+	    else {
+		flags = DATA_FLAGS;    /* gotta do something */
+		error (ERR_NONFATAL, "standard COFF does not support"
+		       " informational sections");
+	    }
+	}
+    }
+
+    for (i=0; i<nsects; i++)
+	if (!strcmp(name, sects[i]->name))
+	    break;
+    if (i == nsects) {
+	if (!strcmp(name, ".text") && !flags)
+	    i = coff_make_section (name, TEXT_FLAGS);
+	else if (!strcmp(name, ".data") && !flags)
+	    i = coff_make_section (name, DATA_FLAGS);
+	else if (!strcmp(name, ".bss") && !flags)
+	    i = coff_make_section (name, BSS_FLAGS);
+	else if (flags)
+	    i = coff_make_section (name, flags);
+	else
+	    i = coff_make_section (name, TEXT_FLAGS);
+	if (flags)
+	    sects[i]->flags = flags;
+    } else if (pass == 1) {
+	if (flags)
+	    error (ERR_WARNING, "section attributes ignored on"
+		   " redeclaration of section `%s'", name);
+    }
+
+    return sects[i]->index;
 }
 
 static void coff_deflabel (char *name, long segment, long offset,
 			   int is_global) {
     int pos = strslen+4;
+    struct Symbol *sym;
 
     if (name[0] == '.' && name[1] == '.') {
 	return;
     }
- 
+
     if (strlen(name) > 8) {
-	if (strslen + 1 + strlen(name) > strsize) {
-	    strsize = strslen + 1 + strlen(name) + STRTAB_DELTA;
-	    strs = nasm_realloc(strs, strsize);
-	}
-	strcpy (strs+strslen, name);
+	saa_wbytes (strs, name, (long)(1+strlen(name)));
 	strslen += 1+strlen(name);
     } else
 	pos = -1;
 
-    if (nsyms >= symsize) {
-	symsize += SYMTAB_DELTA;
-	syms = nasm_realloc(syms, symsize*sizeof(*syms));
-    }
+    sym = saa_wstruct (syms);
 
-    syms[nsyms].strpos = pos;
+    sym->strpos = pos;
     if (pos == -1)
-	strcpy (syms[nsyms].name, name);
-    syms[nsyms].is_global = !!is_global;
+	strcpy (sym->name, name);
+    sym->is_global = !!is_global;
     if (segment == NO_SEG)
-	syms[nsyms].section = -1;      /* absolute symbol */
-    else if (segment == stext.index)
-	syms[nsyms].section = 1;       /* .text */
-    else if (segment == sdata.index)
-	syms[nsyms].section = 2;       /* .data */
-    else if (segment == bssindex)
-	syms[nsyms].section = 3;       /* .bss */
+	sym->section = -1;      /* absolute symbol */
     else {
-	syms[nsyms].section = 0;       /* undefined */
-	syms[nsyms].is_global = TRUE;
+	int i;
+	sym->section = 0;
+	for (i=0; i<nsects; i++)
+	    if (segment == sects[i]->index) {
+		sym->section = i+1;
+		break;
+	    }
+	if (!sym->section)
+	    sym->is_global = TRUE;
     }
     if (is_global == 2)
-	syms[nsyms].value = offset;
+	sym->value = offset;
     else
-	syms[nsyms].value = (syms[nsyms].section == 0 ? 0 : offset);
+	sym->value = (sym->section == 0 ? 0 : offset);
 
     /*
      * define the references from external-symbol segment numbers
      * to these symbol records.
      */
-    if (segment > 2) {
-	if (segment >= bsymsize) {
-	    bsymsize = segment + BSYM_DELTA;
-	    bsym = nasm_realloc (bsym, bsymsize * sizeof(*bsym));
-	}
-	bsym[segment] = nsyms;
-    }
+    if (sym->section == 0)
+	bsym = raa_write (bsym, segment, nsyms);
+
+    if (segment != NO_SEG)
+	symval = raa_write (symval, segment, sym->section ? 0 : sym->value);
 
     nsyms++;
 }
@@ -273,11 +341,20 @@ static long coff_add_reloc (struct Section *sect, long segment,
     r->next = NULL;
 
     r->address = sect->len;
-    r->symbol = (segment == NO_SEG ? SYM_ABS_SEG :
-		 segment == stext.index ? SYM_TEXT_SEG :
-		 segment == sdata.index ? SYM_DATA_SEG :
-		 segment == bssindex ? SYM_BSS_SEG :
-		 bsym[segment] + SYM_INITIAL);
+    if (segment == NO_SEG)
+	r->symbol = 0, r->symbase = ABS_SYMBOL;
+    else {
+	int i;
+	r->symbase = REAL_SYMBOLS;
+	for (i=0; i<nsects; i++)
+	    if (segment == sects[i]->index) {
+		r->symbol = i*2;
+		r->symbase = SECT_SYMBOLS;
+		break;
+	    }
+	if (r->symbase == REAL_SYMBOLS)
+	    r->symbol = raa_read (bsym, segment);
+    }
     r->relative = relative;
 
     sect->nrelocs++;
@@ -285,12 +362,10 @@ static long coff_add_reloc (struct Section *sect, long segment,
     /*
      * Return the fixup for standard COFF common variables.
      */
-    if (r->symbol >= SYM_INITIAL && !win32) {
-	int i = bsym[segment];
-	if (syms[i].section == 0)
-	    return syms[i].value;
-    }
-    return 0;
+    if (r->symbase == REAL_SYMBOLS && !win32)
+	return raa_read (symval, segment);
+    else
+	return 0;
 }
 
 static void coff_out (long segto, void *data, unsigned long type,
@@ -298,6 +373,7 @@ static void coff_out (long segto, void *data, unsigned long type,
     struct Section *s;
     long realbytes = type & OUT_SIZMASK;
     unsigned char mydata[4], *p;
+    int i;
 
     if (wrt != NO_SEG) {
 	wrt = NO_SEG;		       /* continue to do _something_ */
@@ -316,49 +392,53 @@ static void coff_out (long segto, void *data, unsigned long type,
 	return;
     }
 
-    if (segto == stext.index)
-	s = &stext;
-    else if (segto == sdata.index)
-	s = &sdata;
-    else if (segto == bssindex)
-	s = NULL;
-    else {
-	error(ERR_WARNING, "attempt to assemble code in"
-	      " segment %d: defaulting to `.text'", segto);
-	s = &stext;
+    s = NULL;
+    for (i=0; i<nsects; i++)
+	if (segto == sects[i]->index) {
+	    s = sects[i];
+	    break;
+	}
+    if (!s) {
+	int tempint;		       /* ignored */
+	if (segto != coff_section_names (".text", 2, &tempint))
+	    error (ERR_PANIC, "strange segment conditions in COFF driver");
+	else
+	    s = sects[nsects-1];
     }
 
-    if (!s && type != OUT_RESERVE) {
-	error(ERR_WARNING, "attempt to initialise memory in the"
-	      " BSS section: ignored");
+    if (!s->data && type != OUT_RESERVE) {
+	error(ERR_WARNING, "attempt to initialise memory in"
+	      " BSS section `%s': ignored", s->name);
 	if (type == OUT_REL2ADR)
 	    realbytes = 2;
 	else if (type == OUT_REL4ADR)
 	    realbytes = 4;
-	bsslen += realbytes;
+	s->len += realbytes;
 	return;
     }
 
     if (type == OUT_RESERVE) {
-	if (s) {
+	if (s->data) {
 	    error(ERR_WARNING, "uninitialised space declared in"
-		  " %s section: zeroing",
-		  (segto == stext.index ? "code" : "data"));
+		  " non-BSS section `%s': zeroing", s->name);
 	    coff_sect_write (s, NULL, realbytes);
 	} else
-	    bsslen += realbytes;
+	    s->len += realbytes;
     } else if (type == OUT_RAWDATA) {
 	if (segment != NO_SEG)
 	    error(ERR_PANIC, "OUT_RAWDATA with other than NO_SEG");
 	coff_sect_write (s, data, realbytes);
     } else if (type == OUT_ADDRESS) {
-	if (realbytes == 2)
+	if (realbytes == 2 && (segment != NO_SEG || wrt != NO_SEG))
 	    error(ERR_NONFATAL, "COFF format does not support 16-bit"
 		  " relocations");
 	else {
 	    long fix = 0;
-	    if (segment != NO_SEG) {
-		if (segment % 2) {
+	    if (segment != NO_SEG || wrt != NO_SEG) {
+		if (wrt != NO_SEG) {
+		    error(ERR_NONFATAL, "COFF format does not support"
+			  " WRT types");
+		} else if (segment % 2) {
 		    error(ERR_NONFATAL, "COFF format does not support"
 			  " segment base references");
 		} else
@@ -397,14 +477,7 @@ static void coff_out (long segto, void *data, unsigned long type,
 
 static void coff_sect_write (struct Section *sect,
 			     unsigned char *data, unsigned long len) {
-    if (sect->len + len > sect->size) {
-	sect->size = sect->len + len + SECT_DELTA;
-	sect->data = nasm_realloc(sect->data, sect->size);
-    }
-    if (data)
-	memcpy (sect->data+sect->len, data, len);
-    else
-	memset (sect->data+sect->len, 0, len);
+    saa_wbytes (sect->data, data, len);
     sect->len += len;
 }
 
@@ -413,25 +486,35 @@ static int coff_directives (char *directive, char *value, int pass) {
 }
 
 static void coff_write (void) {
-    long textpos, textrelpos, datapos, datarelpos, sympos;
+    long hdrs_end, pos, sympos, vsize;
+    int i;
 
     /*
-     * Work out how big the file will get.
+     * Work out how big the file will get. Calculate the start of
+     * the `real' symbols at the same time.
      */
-    textpos = COFF_HDRS_END;
-    textrelpos = textpos + stext.len;
-    datapos = textrelpos + stext.nrelocs * 10;
-    datarelpos = datapos + sdata.len;
-    sympos = datarelpos + sdata.nrelocs * 10;
+    pos = hdrs_end = 0x14 + 0x28 * nsects;
+    initsym = 3;		       /* two for the file, one absolute */
+    for (i=0; i<nsects; i++) {
+	if (sects[i]->data) {
+	    sects[i]->pos = pos;
+	    pos += sects[i]->len;
+	    sects[i]->relpos = pos;
+	    pos += 10 * sects[i]->nrelocs;
+	} else
+	    sects[i]->pos = sects[i]->relpos = 0L;
+	initsym += 2;		       /* two for each section */
+    }
+    sympos = pos;
 
     /*
      * Output the COFF header.
      */
     fwriteshort (0x14C, coffp);	       /* MACHINE_i386 */
-    fwriteshort (3, coffp);	       /* number of sections */
+    fwriteshort (nsects, coffp);       /* number of sections */
     fwritelong (time(NULL), coffp);    /* time stamp */
     fwritelong (sympos, coffp);
-    fwritelong (nsyms + SYM_INITIAL, coffp);
+    fwritelong (nsyms + initsym, coffp);
     fwriteshort (0, coffp);	       /* no optional header */
     /* Flags: 32-bit, no line numbers. Win32 doesn't even bother with them. */
     fwriteshort (win32 ? 0 : 0x104, coffp);
@@ -439,34 +522,29 @@ static void coff_write (void) {
     /*
      * Output the section headers.
      */
-
-    coff_section_header (".text", 0L, stext.len, textpos,
-			 textrelpos, stext.nrelocs,
-			 (win32 ? 0x60500020L : 0x20L));
-    coff_section_header (".data", stext.len, sdata.len, datapos,
-			 datarelpos, sdata.nrelocs,
-			 (win32 ? 0xC0300040L : 0x40L));
-    coff_section_header (".bss", stext.len+sdata.len, bsslen, 0L, 0L, 0,
-			 (win32 ? 0xC0300080L : 0x80L));
+    vsize = 0L;
+    for (i=0; i<nsects; i++) {
+	coff_section_header (sects[i]->name, vsize, sects[i]->len,
+			     sects[i]->pos, sects[i]->relpos,
+			     sects[i]->nrelocs, sects[i]->flags);
+	vsize += sects[i]->len;
+    }
 
     /*
-     * Output the text section, and its relocations.
+     * Output the sections and their relocations.
      */
-    fwrite (stext.data, stext.len, 1, coffp);
-    coff_write_relocs (&stext);
-
-    /*
-     * Output the data section, and its relocations.
-     */
-    fwrite (sdata.data, sdata.len, 1, coffp);
-    coff_write_relocs (&sdata);
+    for (i=0; i<nsects; i++)
+	if (sects[i]->data) {
+	    saa_fpwrite (sects[i]->data, coffp);
+	    coff_write_relocs (sects[i]);
+	}
 
     /*
      * Output the symbol and string tables.
      */
     coff_write_symbols();
     fwritelong (strslen+4, coffp);     /* length includes length count */
-    fwrite (strs, 1, strslen, coffp);
+    saa_fpwrite (strs, coffp);
 }
 
 static void coff_section_header (char *name, long vsize,
@@ -493,7 +571,9 @@ static void coff_write_relocs (struct Section *s) {
 
     for (r = s->head; r; r = r->next) {
 	fwritelong (r->address, coffp);
-	fwritelong (r->symbol, coffp);
+	fwritelong (r->symbol + (r->symbase == REAL_SYMBOLS ? initsym :
+				 r->symbase == ABS_SYMBOL ? initsym-1 :
+				 r->symbase == SECT_SYMBOLS ? 2 : 0), coffp);
 	/*
 	 * Strange: Microsoft's COFF documentation says 0x03 for an
 	 * absolute relocation, but both Visual C++ and DJGPP agree
@@ -540,17 +620,12 @@ static void coff_write_symbols (void) {
      */
     memset (filename, 0, 18);	       /* useful zeroed buffer */
 
-    coff_symbol (".text", 0L, 0L, 1, 3, 1);
-    fwritelong (stext.len, coffp);
-    fwriteshort (stext.nrelocs, coffp);
-    fwrite (filename, 12, 1, coffp);
-    coff_symbol (".data", 0L, 0L, 2, 3, 1);
-    fwritelong (sdata.len, coffp);
-    fwriteshort (sdata.nrelocs, coffp);
-    fwrite (filename, 12, 1, coffp);
-    coff_symbol (".bss", 0L, 0L, 3, 3, 1);
-    fwritelong (bsslen, coffp);
-    fwrite (filename, 14, 1, coffp);
+    for (i=0; i<nsects; i++) {
+	coff_symbol (sects[i]->name, 0L, 0L, i+1, 3, 1);
+	fwritelong (sects[i]->len, coffp);
+	fwriteshort (sects[i]->nrelocs, coffp);
+	fwrite (filename, 12, 1, coffp);
+    }
 
     /*
      * The absolute symbol, for relative-to-absolute relocations.
@@ -560,10 +635,13 @@ static void coff_write_symbols (void) {
     /*
      * The real symbols.
      */
-    for (i=0; i<nsyms; i++)
-	coff_symbol (syms[i].strpos == -1 ? syms[i].name : NULL,
-		     syms[i].strpos, syms[i].value, syms[i].section,
-		     syms[i].is_global ? 2 : 3, 0);		     
+    saa_rewind (syms);
+    for (i=0; i<nsyms; i++) {
+	struct Symbol *sym = saa_rstruct (syms);
+	coff_symbol (sym->strpos == -1 ? sym->name : NULL,
+		     sym->strpos, sym->value, sym->section,
+		     sym->is_global ? 2 : 3, 0);
+    }
 }
 
 static long coff_segbase (long segment) {
@@ -580,6 +658,10 @@ static void coff_win32_filename (char *inname, char *outname, efunc error) {
     standard_extension (inname, outname, ".obj", error);
 }
 
+#endif /* defined(OF_COFF) || defined(OF_WIN32) */
+
+#ifdef OF_COFF
+
 struct ofmt of_coff = {
     "COFF (i386) object files (e.g. DJGPP for DOS)",
     "coff",
@@ -593,6 +675,10 @@ struct ofmt of_coff = {
     coff_cleanup
 };
 
+#endif
+
+#ifdef OF_WIN32
+
 struct ofmt of_win32 = {
     "Microsoft Win32 (i386) object files",
     "win32",
@@ -605,3 +691,5 @@ struct ofmt of_win32 = {
     coff_win32_filename,
     coff_cleanup
 };
+
+#endif
