@@ -27,9 +27,11 @@ static long reg_flags[] = {	       /* sizes and special flags */
     REG_CREG, REG_CREG, REG_CREG, REG_CR4, REG_CS, REG_CX, REG8,
     REG16, REG8, REG_DREG, REG_DREG, REG_DREG, REG_DREG, REG_DREG,
     REG_DREG, REG_DESS, REG_DX, REG_EAX, REG32, REG32, REG_ECX,
-    REG32, REG32, REG_DESS, REG32, REG32, REG_FSGS, REG_FSGS, REG16,
-    REG16, REG_DESS, FPU0, FPUREG, FPUREG, FPUREG, FPUREG, FPUREG,
-    FPUREG, FPUREG, REG_TREG, REG_TREG, REG_TREG, REG_TREG, REG_TREG
+    REG32, REG32, REG_DESS, REG32, REG32, REG_FSGS, REG_FSGS,
+    MMXREG, MMXREG, MMXREG, MMXREG, MMXREG, MMXREG, MMXREG, MMXREG,
+    REG16, REG16, REG_DESS, FPU0, FPUREG, FPUREG, FPUREG, FPUREG,
+    FPUREG, FPUREG, FPUREG, REG_TREG, REG_TREG, REG_TREG, REG_TREG,
+    REG_TREG
 };
 
 enum {				       /* special tokens */
@@ -43,7 +45,8 @@ static char *special_names[] = {       /* and the actual text */
 };
 
 static char *prefix_names[] = {
-    "lock", "rep", "repe", "repne", "repnz", "repz", "times"
+    "a16", "a32", "lock", "o16", "o32", "rep", "repe", "repne",
+    "repnz", "repz", "times"
 };
 
 /*
@@ -110,12 +113,15 @@ static struct ofmt *outfmt;
 
 static long seg, ofs;
 
+static int forward;
+
 insn *parse_line (long segment, long offset, lfunc lookup_label, int pass,
 		  char *buffer, insn *result, struct ofmt *output,
 		  efunc errfunc) {
     int operand;
     int critical;
 
+    forward = result->forw_ref = FALSE;
     q = tempstorage;
     bufptr = buffer;
     labelfunc = lookup_label;
@@ -182,9 +188,9 @@ insn *parse_line (long segment, long offset, lfunc lookup_label, int pass,
 	    } else
 		result->times = value->value;
 	} else {
-	    if (result->nprefix == 3)
+	    if (result->nprefix == MAXPREFIX)
 		error (ERR_NONFATAL,
-		       "instruction has more than three prefixes");
+		       "instruction has more than %d prefixes", MAXPREFIX);
 	    else
 		result->prefixes[result->nprefix++] = tokval.t_integer;
 	    i = nexttoken();
@@ -207,7 +213,7 @@ insn *parse_line (long segment, long offset, lfunc lookup_label, int pass,
      * `critical' flag on calling evaluate(), so that it will bomb
      * out on undefined symbols. Nasty, but there's nothing we can
      * do about it.
-     * 
+     *
      * For the moment, EQU has the same difficulty, so we'll
      * include that.
      */
@@ -247,27 +253,48 @@ insn *parse_line (long segment, long offset, lfunc lookup_label, int pass,
 		eop->stringval = tokval.t_charptr;
 		eop->stringlen = tokval.t_inttwo;
 		i = nexttoken();       /* eat the comma */
-	    } else if (i == TOKEN_FLOAT) {
-		eop->type = EOT_DB_STRING;
-		eop->stringval = q;
-		if (result->opcode == I_DD)
-		    eop->stringlen = 4;
-		else if (result->opcode == I_DQ)
-		    eop->stringlen = 8;
-		else if (result->opcode == I_DT)
-		    eop->stringlen = 10;
-		else {
-		    error(ERR_NONFATAL, "floating-point constant encountered"
-			  " in `D%c' instruction",
-			  result->opcode == I_DW ? 'W' : 'B');
-		    eop->type = EOT_NOTHING;
+		continue;
+	    }
+
+	    if (i == TOKEN_FLOAT || i == '-') {
+		long sign = +1L;
+
+		if (i == '-') {
+		    char *save = bufptr;
+		    i = nexttoken();
+		    sign = -1L;
+		    if (i != TOKEN_FLOAT) {
+			bufptr = save;
+			i = '-';
+		    }
 		}
-		q += eop->stringlen;
-		if (!float_const (tokval.t_charptr,
-				  (unsigned char *)eop->stringval,
-				  eop->stringlen, error))
-		    eop->type = EOT_NOTHING;
-	    } else {
+
+		if (i == TOKEN_FLOAT) {
+		    eop->type = EOT_DB_STRING;
+		    eop->stringval = q;
+		    if (result->opcode == I_DD)
+			eop->stringlen = 4;
+		    else if (result->opcode == I_DQ)
+			eop->stringlen = 8;
+		    else if (result->opcode == I_DT)
+		    eop->stringlen = 10;
+		    else {
+			error(ERR_NONFATAL, "floating-point constant"
+			      " encountered in `D%c' instruction",
+			      result->opcode == I_DW ? 'W' : 'B');
+			eop->type = EOT_NOTHING;
+		    }
+		    q += eop->stringlen;
+		    if (!float_const (tokval.t_charptr, sign,
+				      (unsigned char *)eop->stringval,
+				      eop->stringlen, error))
+			eop->type = EOT_NOTHING;
+		    i = nexttoken();       /* eat the comma */
+		    continue;
+		}
+	    }
+
+	    /* anything else */ {
 		expr *value;
 		eval_reset();
 		value = evaluate (critical);
@@ -289,7 +316,7 @@ insn *parse_line (long segment, long offset, lfunc lookup_label, int pass,
 	    }
 	}
 	return result;
-    }	
+    }
 
     /* right. Now we begin to parse the operands. There may be up to three
      * of these, separated by commas, and terminated by a zero token. */
@@ -360,6 +387,8 @@ insn *parse_line (long segment, long offset, lfunc lookup_label, int pass,
 	eval_reset();
 
 	value = evaluate (critical);
+	if (forward)
+	    result->forw_ref = TRUE;
 	if (!value) {		       /* error in evaluator */
 	    result->opcode = -1;       /* unrecoverable parse error: */
 	    return result;	       /* ignore this instruction */
@@ -383,6 +412,8 @@ insn *parse_line (long segment, long offset, lfunc lookup_label, int pass,
 		i = nexttoken();
 	    }
 	    value = evaluate (critical);
+	    if (forward)
+		result->forw_ref = TRUE;	
 	    /* and get the offset */
 	    if (!value) {	       /* but, error in evaluator */
 		result->opcode = -1;   /* unrecoverable parse error: */
@@ -420,9 +451,10 @@ insn *parse_line (long segment, long offset, lfunc lookup_label, int pass,
 		if (seg[1].type!=0 || seg->value!=1 ||
 		    REG_SREG & ~reg_flags[seg->type])
 		    error (ERR_NONFATAL, "invalid segment override");
-		else if (result->nprefix == 3)
+		else if (result->nprefix == MAXPREFIX)
 		    error (ERR_NONFATAL,
-			   "instruction has more than three prefixes");
+			   "instruction has more than %d prefixes",
+			   MAXPREFIX);
 		else
 		    result->prefixes[result->nprefix++] = seg->type;
 	    }
@@ -467,20 +499,40 @@ insn *parse_line (long segment, long offset, lfunc lookup_label, int pass,
 			e++;
 		    } else
 			result->oprs[operand].wrt = NO_SEG;
-		    if (e->type != 0) {   /* is there a segment id? */
-			if (e->type < EXPR_SEGBASE) {
-			    error (ERR_NONFATAL,
-				   "invalid effective address");
-			    result->opcode = -1;
-			    return result;
-			} else
-			    result->oprs[operand].segment = (e->type -
-							     EXPR_SEGBASE);
+		    /*
+		     * Look for a segment base type.
+		     */
+		    if (e->type && e->type < EXPR_SEGBASE) {
+			error (ERR_NONFATAL, "invalid effective address");
+			result->opcode = -1;
+			return result;
+		    }
+		    while (e->type && e->value == 0)
+			e++;
+		    if (e->type && e->value != 1) {
+			error (ERR_NONFATAL, "invalid effective address");
+			result->opcode = -1;
+			return result;
+		    }
+		    if (e->type) {
+			result->oprs[operand].segment = e->type-EXPR_SEGBASE;
 			e++;
 		    } else
 			result->oprs[operand].segment = NO_SEG;
+		    while (e->type && e->value == 0)
+			e++;
+		    if (e->type) {
+			error (ERR_NONFATAL, "invalid effective address");
+			result->opcode = -1;
+			return result;
+		    }
 		}
+	    } else {
+		o = 0;
+		result->oprs[operand].wrt = NO_SEG;
+		result->oprs[operand].segment = NO_SEG;
 	    }
+
 	    if (e->type != 0) {    /* there'd better be nothing left! */
 		error (ERR_NONFATAL, "invalid effective address");
 		result->opcode = -1;
@@ -549,8 +601,9 @@ static int is_comma_next (void) {
  * (for local labels), whereas a number may appear anywhere *but* at the
  * start. */
 
-#define isidstart(c) ( isalpha(c) || (c)=='_' || (c)=='.' )
-#define isidchar(c)  ( isalnum(c) || (c)=='_' || (c)=='$' || (c)=='.' )
+#define isidstart(c) ( isalpha(c) || (c)=='_' || (c)=='.' || (c)=='?' )
+#define isidchar(c)  ( isidstart(c) || isdigit(c) || (c)=='$' || (c)=='#' \
+                                                  || (c)=='@' || (c)=='~' )
 
 /* Ditto for numeric constants. */
 
@@ -561,6 +614,12 @@ static int is_comma_next (void) {
 
 #define numvalue(c)  ((c)>='a' ? (c)-'a'+10 : (c)>='A' ? (c)-'A'+10 : (c)-'0')
 
+/*
+ * This tokeniser routine has only one side effect, that of
+ * updating `bufptr'. Hence by saving `bufptr', lookahead may be
+ * performed.
+ */
+
 static int nexttoken (void) {
     char ourcopy[256], *r, *s;
 
@@ -568,8 +627,16 @@ static int nexttoken (void) {
     if (!*bufptr) return 0;
 
     /* we have a token; either an id, a number or a char */
-    if (isidstart(*bufptr)) {	       /* now we've got an identifier */
+    if (isidstart(*bufptr) ||
+	(*bufptr == '$' && isidstart(bufptr[1]))) {
+	/* now we've got an identifier */
 	int i;
+	int is_sym = FALSE;
+
+	if (*bufptr == '$') {
+	    is_sym = TRUE;
+	    bufptr++;
+	}
 
  	tokval.t_charptr = q;
 	*q++ = *bufptr++;
@@ -578,6 +645,8 @@ static int nexttoken (void) {
 	for (s=tokval.t_charptr, r=ourcopy; *s; s++)
 	    *r++ = tolower (*s);
 	*r = '\0';
+	if (is_sym)
+	    return TOKEN_ID;	       /* bypass all other checks */
 	/* right, so we have an identifier sitting in temp storage. now,
 	 * is it actually a register or instruction name, or what? */
 	if ((tokval.t_integer=bsi(ourcopy, reg_names,
@@ -916,7 +985,7 @@ static expr *segment_part (expr *e) {
  * which is TRUE if the evaluation is critical (i.e. unresolved
  * symbols are an error condition). Must update the global `i' to
  * reflect the token after the parsed string. May return NULL.
- * 
+ *
  * evaluate() should report its own errors: on return it is assumed
  * that if NULL has been returned, the error has already been
  * reported.
@@ -924,7 +993,7 @@ static expr *segment_part (expr *e) {
 
 /*
  * Grammar parsed is:
- * 
+ *
  * expr  : expr0 [ WRT expr6 ]
  * expr0 : expr1 [ {|} expr1]
  * expr1 : expr2 [ {^} expr2]
@@ -988,7 +1057,7 @@ static expr *expr2(int critical) {
     e = expr3(critical);
     if (!e)
 	return NULL;
-    while (i == '&') {	
+    while (i == '&') {
 	i = nexttoken();
 	f = expr3(critical);
 	if (!f)
@@ -1170,9 +1239,9 @@ static expr *expr6(int critical) {
 	     * Since the whole line is parsed before the label it
 	     * defines is given to the label manager, we have
 	     * problems with lines such as
-	     * 
+	     *
 	     *   end: TIMES 512-(end-start) DB 0
-	     * 
+	     *
 	     * where `end' is not known on pass one, despite not
 	     * really being a forward reference, and due to
 	     * criticality it is _needed_. Hence we check our label
@@ -1195,6 +1264,7 @@ static expr *expr6(int critical) {
 			   tokval.t_charptr);
 		    return NULL;
 		} else {
+		    forward = TRUE;
 		    label_seg = seg;
 		    label_ofs = ofs;
 		}
